@@ -7,7 +7,7 @@ use crate::{
     capped_hashset::CappedHashSet,
     exomind::{email_trait_id, thread_entity_id, ExomindClient, ExomindHistoryAction},
     gmail::{GmailAccount, GmailClient, GmailHistoryAction, HistoryId},
-    parsing,
+    parsing::{self, FlaggedEmail},
 };
 use exocore::{
     core::time::{ConsistentTimestamp, Utc},
@@ -20,7 +20,7 @@ use exocore::{
         mutation::OperationId,
     },
 };
-use exomind_protos::base::{CollectionChild, Email, EmailThread};
+use exomind_protos::base::{CollectionChild, Email, EmailThread, Unread};
 
 pub struct AccountSynchronizer {
     pub account: GmailAccount,
@@ -97,16 +97,9 @@ impl AccountSynchronizer {
 
         // move to inbox emails that are in exomind's inbox
         for (thread_id, thread) in exomind_threads {
-            self.update_last_exomind_operation(thread.operation_id);
-
             if thread.account_entity_id == self.account.entity_id && !in_gmail.contains(&thread_id)
             {
-                let history_id = thread.add_to_gmail_inbox(&self.gmail).await?;
-                self.update_last_gmail_history(history_id);
-
-                if let Some(history_id) = history_id {
-                    self.generated_gmail_history.insert(history_id);
-                }
+                self.add_to_gmail_inbox(thread.thread_id()).await?;
             }
         }
 
@@ -147,18 +140,18 @@ impl AccountSynchronizer {
         }
 
         for history in history_list {
+            let history_id = history.history_id();
+            self.update_last_gmail_history(Some(history_id));
+            if self.generated_gmail_history.contains(&history_id) {
+                debug!(
+                    "Dropping history {} because generated by ourself",
+                    history_id
+                );
+                continue;
+            }
+
             match history {
-                GmailHistoryAction::AddToInbox(history_id, thread) => {
-                    self.update_last_gmail_history(Some(history_id));
-
-                    if self.generated_gmail_history.contains(&history_id) {
-                        info!(
-                            "Dropping history {} because generated by ourself",
-                            history_id
-                        );
-                        continue;
-                    }
-
+                GmailHistoryAction::AddToInbox(_history_id, thread) => {
                     let sync_thread = SynchronizedThread::from_gmail(self.account.clone(), thread);
                     if let Some(sync_thread) = sync_thread {
                         let current_entity = self
@@ -174,20 +167,18 @@ impl AccountSynchronizer {
                         self.generated_exomind_operations.insert_all(&operation_ids);
                     }
                 }
-                GmailHistoryAction::RemoveFromInbox(history_id, thread_id) => {
-                    self.update_last_gmail_history(Some(history_id));
-
-                    if self.generated_gmail_history.contains(&history_id) {
-                        info!(
-                            "Dropping history {} because generated by ourself",
-                            history_id
-                        );
-                        continue;
-                    }
-
+                GmailHistoryAction::RemoveFromInbox(_history_id, thread_id) => {
                     let operation_ids = self.exomind.remove_from_inbox(&thread_id).await?;
                     self.generated_exomind_operations.insert_all(&operation_ids);
                 }
+                GmailHistoryAction::MarkUnread(_history_id, thread_id, msg_id) => {
+                    let operation_ids = self.exomind.mark_email_unread(&thread_id, &msg_id).await?;
+                    self.generated_exomind_operations.insert_all(&operation_ids);
+                }
+                GmailHistoryAction::MarkRead(_history_id, thread_id, msg_id) => {
+                    let operation_ids = self.exomind.mark_email_read(&thread_id, &msg_id).await?;
+                    self.generated_exomind_operations.insert_all(&operation_ids);
+                }
             }
         }
 
@@ -195,55 +186,53 @@ impl AccountSynchronizer {
     }
 
     pub async fn synchronize_exomind_history(&mut self) -> anyhow::Result<()> {
-        let last_operation_id = self.last_exomind_operation.unwrap_or_default();
+        let last_operation_id = if let Some(last_operation) = self.last_exomind_operation {
+            last_operation
+        } else {
+            // never fetched through history, we fetch last exomind history action for next sync via history
+            if let Some(last_action) = self.latest_exomind_history_action().await? {
+                self.last_exomind_operation = Some(last_action.operation_id());
+            }
+            return Ok(());
+        };
 
-        let history_list = self
+        let history_actions = self
             .exomind
             .list_inbox_history(&self.account, last_operation_id)
             .await?;
 
-        if !history_list.is_empty() {
+        if !history_actions.is_empty() {
             info!(
                 "Fetched {} history from exomind after operation {} ({:?})",
-                history_list.len(),
+                history_actions.len(),
                 last_operation_id,
                 ConsistentTimestamp::from(last_operation_id).to_datetime(),
             );
         }
 
-        for history in history_list {
-            match history {
-                ExomindHistoryAction::AddToInbox(operation_id, thread) => {
-                    self.update_last_exomind_operation(thread.operation_id);
+        for action in history_actions {
+            let operation_id = action.operation_id();
+            self.update_last_exomind_operation(Some(operation_id));
+            if self.generated_exomind_operations.contains(&operation_id) {
+                debug!(
+                    "Dropping operation {} because generated by ourself",
+                    operation_id
+                );
+                continue;
+            }
 
-                    if self.generated_exomind_operations.contains(&operation_id) {
-                        info!(
-                            "Dropping operation {} because generated by ourself",
-                            operation_id
-                        );
-                        continue;
-                    }
-
-                    let history_id = thread.add_to_gmail_inbox(&self.gmail).await?;
-                    if let Some(history_id) = history_id {
-                        self.generated_gmail_history.insert(history_id);
-                    }
+            match action {
+                ExomindHistoryAction::AddToInbox(_operation_id, thread) => {
+                    self.add_to_gmail_inbox(thread.thread_id()).await?;
                 }
-                ExomindHistoryAction::RemovedFromInbox(operation_id, thread) => {
-                    self.update_last_exomind_operation(thread.operation_id);
-
-                    if self.generated_exomind_operations.contains(&operation_id) {
-                        info!(
-                            "Dropping operation {} because generated by ourself",
-                            operation_id
-                        );
-                        continue;
-                    }
-
-                    let history_id = thread.remove_from_gmail_inbox(&self.gmail).await?;
-                    if let Some(history_id) = history_id {
-                        self.generated_gmail_history.insert(history_id);
-                    }
+                ExomindHistoryAction::RemovedFromInbox(_operation_id, thread) => {
+                    self.remove_from_gmail_inbox(thread.thread_id()).await?;
+                }
+                ExomindHistoryAction::MarkRead(_operation_id, message_id) => {
+                    self.mark_gmail_read(&message_id).await?;
+                }
+                ExomindHistoryAction::MarkUnread(_, message_id) => {
+                    self.mark_gmail_unread(&message_id).await?;
                 }
             }
         }
@@ -251,6 +240,13 @@ impl AccountSynchronizer {
         Ok(())
     }
 
+    pub async fn latest_exomind_history_action(
+        &mut self,
+    ) -> anyhow::Result<Option<ExomindHistoryAction>> {
+        let history_list = self.exomind.list_inbox_history(&self.account, 0).await?;
+        Ok(history_list.into_iter().next())
+    }
+
     pub fn update_last_gmail_history(&mut self, history_id: Option<HistoryId>) {
         let history_id = if let Some(history_id) = history_id {
             history_id
@@ -286,13 +282,77 @@ impl AccountSynchronizer {
             _ => {}
         }
     }
+
+    async fn add_to_gmail_inbox(&mut self, thread_id: &str) -> anyhow::Result<()> {
+        let thread = self
+            .gmail
+            .add_thread_label(thread_id, "INBOX".to_string())
+            .await?;
+
+        let history_id = thread
+            .history_id
+            .and_then(|history| history.parse::<HistoryId>().ok());
+        if let Some(history_id) = history_id {
+            self.generated_gmail_history.insert(history_id);
+        }
+
+        Ok(())
+    }
+
+    async fn remove_from_gmail_inbox(&mut self, thread_id: &str) -> anyhow::Result<()> {
+        let thread = self
+            .gmail
+            .remove_thread_label(thread_id, "INBOX".to_string())
+            .await?;
+
+        let history_id = thread
+            .history_id
+            .and_then(|history| history.parse::<HistoryId>().ok());
+        if let Some(history_id) = history_id {
+            self.generated_gmail_history.insert(history_id);
+        }
+
+        Ok(())
+    }
+
+    async fn mark_gmail_unread(&mut self, message_id: &str) -> anyhow::Result<()> {
+        let message = self
+            .gmail
+            .add_message_label(message_id, "UNREAD".to_string())
+            .await?;
+
+        let history_id = message
+            .history_id
+            .and_then(|history| history.parse::<HistoryId>().ok());
+        if let Some(history_id) = history_id {
+            self.generated_gmail_history.insert(history_id);
+        }
+
+        Ok(())
+    }
+
+    async fn mark_gmail_read(&mut self, message_id: &str) -> anyhow::Result<()> {
+        let message = self
+            .gmail
+            .remove_message_label(message_id, "UNREAD".to_string())
+            .await?;
+
+        let history_id = message
+            .history_id
+            .and_then(|history| history.parse::<HistoryId>().ok());
+        if let Some(history_id) = history_id {
+            self.generated_gmail_history.insert(history_id);
+        }
+
+        Ok(())
+    }
 }
 
 pub struct SynchronizedThread {
     pub account_entity_id: String,
     pub thread: EmailThread,
-    pub emails: Vec<Email>,
-    pub _inbox_child: Option<CollectionChild>,
+    pub emails: Vec<FlaggedEmail>,
+    pub inbox_child: Option<CollectionChild>,
     pub history_id: Option<HistoryId>,
     pub operation_id: Option<OperationId>,
 }
@@ -308,17 +368,29 @@ impl SynchronizedThread {
             .find(|c| c.instance.collection.as_ref().map(|c| c.entity_id.as_ref()) == Some("inbox"))
             .map(|t| t.instance);
 
-        let emails: Vec<Email> = entity
+        let email_unread_flags = entity
+            .traits_of_type::<Unread>()
+            .into_iter()
+            .flat_map(|flag| flag.instance.entity.map(|refer| refer.trait_id))
+            .collect::<HashSet<_>>();
+
+        let emails: Vec<FlaggedEmail> = entity
             .traits_of_type::<Email>()
             .into_iter()
-            .map(|t| t.instance)
+            .map(|t| {
+                let unread = email_unread_flags.contains(&email_trait_id(&t.instance.source_id));
+                FlaggedEmail {
+                    proto: t.instance,
+                    unread,
+                }
+            })
             .collect();
 
         Some(SynchronizedThread {
             account_entity_id,
             thread,
             emails,
-            _inbox_child: inbox_child,
+            inbox_child,
             history_id: None,
             operation_id: Some(entity.last_operation_id),
         })
@@ -354,21 +426,25 @@ impl SynchronizedThread {
 
         thread.account = Some(account_ref.clone());
         for email in &mut emails {
-            email.account = Some(account_ref.clone());
+            email.proto.account = Some(account_ref.clone());
         }
 
         if let Some(email) = emails.last() {
-            thread.snippet = email.snippet.clone();
-            thread.subject = email.subject.clone();
-            thread.from = email.from.clone();
+            thread.snippet = email.proto.snippet.clone();
+            thread.subject = email.proto.subject.clone();
+            thread.from = email.proto.from.clone();
             thread.last_email = Some(Reference {
                 entity_id: thread_entity_id,
-                trait_id: email_trait_id(email),
+                trait_id: email_trait_id(&email.proto.source_id),
             })
         }
 
-        let thread_create_date = emails.first().and_then(|email| email.received_date.clone());
-        let thread_modification_date = emails.last().and_then(|email| email.received_date.clone());
+        let thread_create_date = emails
+            .first()
+            .and_then(|email| email.proto.received_date.clone());
+        let thread_modification_date = emails
+            .last()
+            .and_then(|email| email.proto.received_date.clone());
         let thread_last_date = thread_modification_date
             .as_ref()
             .or_else(|| thread_create_date.as_ref())
@@ -391,7 +467,7 @@ impl SynchronizedThread {
             account_entity_id: account.entity_id,
             thread,
             emails,
-            _inbox_child: inbox_child,
+            inbox_child,
             history_id,
             operation_id: None,
         })
@@ -404,25 +480,4 @@ impl SynchronizedThread {
     pub fn entity_id(&self) -> EntityId {
         thread_entity_id(&self.thread)
     }
-
-    pub async fn add_to_gmail_inbox(&self, gmc: &GmailClient) -> anyhow::Result<Option<HistoryId>> {
-        let thread = gmc.add_label(self.thread_id(), "INBOX".to_string()).await?;
-        let history_id = thread
-            .history_id
-            .and_then(|history| history.parse::<HistoryId>().ok());
-        Ok(history_id)
-    }
-
-    pub async fn remove_from_gmail_inbox(
-        &self,
-        gmc: &GmailClient,
-    ) -> anyhow::Result<Option<HistoryId>> {
-        let thread = gmc
-            .remove_label(self.thread_id(), "INBOX".to_string())
-            .await?;
-        let history_id = thread
-            .history_id
-            .and_then(|history| history.parse::<HistoryId>().ok());
-        Ok(history_id)
-    }
 }
